"""
Minotaur. A python, asynchronous, inotify interface.
"""

from .__version__ import \
    __title__, \
    __description__, \
    __url__, \
    __author__, \
    __author_email__, \
    __copyright__, \
    __license__, \
    __version__

from ._mask import Mask
from ._base import Inotify, InotifyBase
from ._event import Event
from ._minotaur import Minotaur
from . import _inotify

init = _inotify.init
add_watch = _inotify.add_watch
rm_watch = _inotify.rm_watch

__all__ = (
    'Mask',
    'Inotify',
    'InotifyBase',
    'Minotaur',
    'Event',

    'init',
    'add_watch',
    'rm_watch',
)



from typing import Deque, Optional
from pathlib import Path
from collections import deque
from io import DEFAULT_BUFFER_SIZE
from struct import Struct
import os as _os

import asyncio

from . import _inotify
from ._mask import Mask
from ._event import Event

__all__ = ('InotifyBase', 'Inotify')


class InotifyBase:
    __slots__ = (
        '_nonblock',
        '_cloexec',
        '_fd',
        '_buf',
        '_loop',
        '_waitq',
    )

    _loop: Optional[asyncio.AbstractEventLoop]

    _event = Struct('=iIII')
    _event_sz = _event.size

    def __init__(self,
                 blocking=True,
                 cloexec=True,
                 loop: Optional[asyncio.AbstractEventLoop] = None):

        self._nonblock = not blocking
        self._cloexec = cloexec
        self._fd = -1

        if self._nonblock:
            if loop is None:
                loop = asyncio.get_event_loop()
            self._loop = loop
            self._waitq: Deque[asyncio.Future] = deque()
        else:
            self._loop = None

    @property
    def closed(self) -> bool:
        """
        True if the inotify fd is closed
        """
        return self._fd < 0

    def open(self):
        """
        Create the inotify fd
        """

        assert(self._fd < 0)
        nb = self._nonblock and _inotify.IN_NONBLOCK
        ce = self._cloexec and _inotify.IN_CLOEXEC
        self._fd = _inotify.init(nb | ce)

    def _check_open(self):
        if self.closed:
            raise ValueError("I/O operation on closed file.")

    def _register_for_read(self):
        self._check_open()
        assert(self._nonblock)
        assert(self._loop is not None)
        self._loop.add_reader(self._fd, self._fd_readable)

    def _unregister_for_read(self):
        self._check_open()
        assert(self._nonblock)
        assert(self._loop is not None)
        self._loop.remove_reader(self._fd)

    def _fd_readable(self):
        """
        Callback for asyncio when the inotify fd has become readable
        """

        while self._waitq:
            w = self._waitq.popleft()
            if not self._waitq:
                self._unregister_for_read()
            if not w.done():
                w.set_result(None)
                break

    def close(self):
        """
        Close the inotify fd

        This will wake all asynchronous waiters.
        """

        if not self.closed:
            try:
                _os.close(self._fd)
            finally:
                self._fd = -1

        # Must wake all waiters to let them exit their loops now that we're
        # closed
        if self._loop:
            assert(self._loop is not None)
            try:
                self._loop.remove_reader(self._fd)
            except ValueError:
                pass
            while self._waitq:
                w = self._waitq.popleft()
                if not w.done():
                    w.set_result(None)

    def __del__(self):
        self.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()

    def add_watch(self, p: Path, mask: Mask) -> int:
        """
        Add a watch to the inotify fd.

        Return the watch descriptor (an integer).
        """

        self._check_open()
        path = _os.fsencode(p)
        return _inotify.add_watch(self._fd, path, mask.value)

    def rm_watch(self, wd: int) -> int:
        """
        Remove a watch from the inotify fd, given a watch descriptor
        """

        self._check_open()
        return _inotify.rm_watch(self._fd, wd)

    def read(self, size: Optional[int] = None) -> bytes:
        """
        Read raw bytes from the inotify fd.

        This is probably not the interface that you want. Use the iterator
        protocol instead.
        """

        if size is None or size < 0:
            return self.readall()

        self._check_open()
        return _os.read(self._fd, size)

    def readall(self) -> bytes:
        """
        Read all raw bytes from the inotify fd.

        This only works in non-blocking mode, since in blocking mode the
        returned object would, by definition, be infinite in size.
        """

        self._check_open()

        result = bytearray()
        n = DEFAULT_BUFFER_SIZE

        if not self._nonblock:
            raise ValueError("Can't readall on blocking inotify")

        while True:
            try:
                chunk = _os.read(self._fd, n)
            except BlockingIOError:
                if result:
                    break
                return b''
            if not chunk:
                break
            result += chunk

        return bytes(result)

    def __iter__(self):
        self._check_open()

        if self._nonblock:
            raise ValueError("Synchronous iiter used on non-blocking inotify")

        if hasattr(self, '_buf'):
            raise ValueError("Concurrent iteration on inotify fd")

        self._buf = b''
        return self

    def __aiter__(self):
        self._check_open()

        if not self._nonblock:
            raise ValueError("Async iiter used on blocking inotify")

        if not self._loop:
            raise ValueError("No event loop for async iter")

        if hasattr(self, '_buf'):
            raise ValueError("Concurrent iteration on inotify fd")

        self._buf = b''
        return self

    def _fill_buffer(self, chunk_size: int = DEFAULT_BUFFER_SIZE) -> bool:
        """
        Read chunk_size bytes from the fd and append to the internal buffer.
        """

        chunk = self.read(chunk_size)
        if self._buf:
            self._buf += chunk
        else:
            self._buf = chunk
        return bool(chunk)

    def _buf_pop(self) -> Optional[Event]:
        """
        Parse the first event frm the buffer and return it as an Event object.
        If there is not enough data in the buffer then return None.
        """

        hdr_len = self._event_sz

        if len(self._buf) < hdr_len:
            return None

        wd, mask, cookie, evlen = self._event.unpack_from(self._buf)

        tot_len = hdr_len + evlen
        if len(self._buf) < tot_len:
            return None

        path_buf = self._buf[hdr_len:tot_len]
        try:
            nul = path_buf.find(0)
        except ValueError:
            pass
        else:
            path_buf = path_buf[:nul]

        path = _os.fsdecode(path_buf)
        self._buf = self._buf[tot_len:]

        return Event(wd, Mask(mask), cookie, Path(path))

    def _next_event(self) -> Optional[Event]:
        while not self.closed:
            ret = self._buf_pop()
            if ret:
                return ret

            self._fill_buffer()

        return None

    async def _next_event_async(self) -> Optional[Event]:
        while not self.closed:
            ret = self._buf_pop()
            if ret:
                return ret

            if not self._fill_buffer(-1):
                assert(self._loop is not None)

                if not self._waitq:
                    self._register_for_read()
                wake = self._loop.create_future()
                self._waitq.append(wake)
                await wake

        return None


class Inotify(InotifyBase):
    def __next__(self) -> Event:
        evt = self._next_event()
        if evt is not None:
            return evt
        raise StopIteration

    async def __anext__(self) -> Event:
        evt = await self._next_event_async()
        if evt is not None:
            return evt
        raise StopIteration



from enum import IntFlag as _IntFlag

from . import _inotify


__all__ = ('Mask',)


class Mask(_IntFlag):
    show_help: bool

    def __new__(cls, value, doc=None, show_help=True):
        # int.__new__ needs a stub in the typeshed
        # https://github.com/python/typeshed/issues/2686
        #
        # but that broke something else, so they removed it
        # https://github.com/python/typeshed/issues/1464
        #
        # We have no choice but to ignore mypy error here :(
        self = int.__new__(cls, value)  # type: ignore
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
            self.show_help = show_help
        return self

    """
    Flags for establishing inotify watches.
    """

    ACCESS = _inotify.IN_ACCESS, 'File was accessed'
    ATTRIB = _inotify.IN_ATTRIB, 'Metaata changed, eg. permissions'
    CLOSE_WRITE = _inotify.IN_CLOSE_WRITE, 'File for writing was closed'
    CLOSE_NOWRITE = _inotify.IN_CLOSE_NOWRITE, \
        'File or dir not opened for writing was closed'
    CREATE = _inotify.IN_CREATE, 'File/dir was created'
    DELETE = _inotify.IN_DELETE, 'File or dir was deleted'
    DELETE_SELF = _inotify.IN_DELETE_SELF, \
        'Watched file/dir was itself deleted'
    MODIFY = _inotify.IN_MODIFY, 'File was modified'
    MOVE_SELF = _inotify.IN_MOVE_SELF, 'Watched file/dir was itself moved'
    MOVED_FROM = _inotify.IN_MOVED_FROM, \
        'Generated for dir containing old filename when a file is renamed'
    MOVED_TO = _inotify.IN_MOVED_TO, \
        'Generated for dir containing new filename when a file is renamed'
    OPEN = _inotify.IN_OPEN, 'File or dir was opened'
    MOVE = _inotify.IN_MOVE, 'MOVED_FROM | MOVED_TO'
    CLOSE = _inotify.IN_CLOSE, 'IN_CLOSE_WRITE | IN_CLOSE_NOWRITE'

    DONT_FOLLOW = _inotify.IN_DONT_FOLLOW, \
        "Don't dereference pathname if it is a symbolic link"
    EXCL_UNLINK = _inotify.IN_EXCL_UNLINK, \
        "Don't generate events after files have been unlinked"

    MASK_ADD = _inotify.IN_MASK_ADD, 'Add flags to an existing watch', False

    ONESHOT = _inotify.IN_ONESHOT, 'Only generate one event for this watch'
    ONLYDIR = _inotify.IN_ONLYDIR, 'Watch pathname only if it is a dir'
    MASK_CREATE = _inotify.IN_MASK_CREATE, \
        "Only watch path if it isn't already being watched"

    # These are returned in events
    IGNORED = _inotify.IN_IGNORED, 'Watch was removed', False
    ISDIR = _inotify.IN_ISDIR, 'This event is a dir', False
    Q_OVERFLOW = _inotify.IN_Q_OVERFLOW, 'Event queue overflowed', False
    UNMOUNT = _inotify.IN_UNMOUNT, \
        'Filesystem containing watched object was unmounted', False

    EVENT_TYPE = _inotify.EVENT_TYPE_MASK, 'Mask of all event types', False



from typing import Sequence
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from functools import reduce
from operator import or_
import asyncio
import re

from minotaur import Inotify, Mask, Minotaur
from minotaur import __title__


_splitter = re.compile(r'[,\.|:\-\s]+')


def _sync_main(cls, dirs: Sequence[Path], mask: Mask):
    with cls() as n:
        for p in dirs:
            wd = n.add_watch(p, mask)
            print(f'{p.resolve()}: wd={wd}')

        for evt in n:
            print(evt)


async def _task(cls, dirs: Sequence[Path], mask: Mask):
    with cls(blocking=False) as n:
        for p in dirs:
            wd = n.add_watch(p, mask)
            print(f'{p.resolve()}: wd={wd}')

        async for evt in n:
            print(evt)


def _async_main(cls, dirs: Sequence[Path], mask: Mask):
    loop = asyncio.get_event_loop()
    coro = _task(cls, dirs, mask)
    task = loop.create_task(coro)
    loop.run_until_complete(task)


_mask_help = '\n'.join((f' - {x.name}: {x.__doc__}'
                        for x in Mask if x.show_help))

_prog_desc = f"""
Minotaur: A pythonic, asynchronous, inotify interface.

A summary of inotify watch flags:
{_mask_help}
"""

_default_mask = 'create,delete,delete_self,moved_from,moved_to'


def _main():
    opts = ArgumentParser(prog=__title__,
                          formatter_class=RawTextHelpFormatter,
                          description=_prog_desc)

    opts.set_defaults(func=_async_main, cls=Inotify)

    g = opts.add_mutually_exclusive_group(required=False)
    g.add_argument('--async', '-a',
                   dest='func',
                   action='store_const',
                   const=_async_main,
                   help='Use asyncio event loop')
    g.add_argument('--sync', '-s',
                   dest='func',
                   action='store_const',
                   const=_sync_main,
                   help='Use synchronous interface')

    opts.add_argument('--fancy', '-f',
                      dest='cls',
                      action='store_const',
                      const=Minotaur,
                      help='Use fancy interface')
    opts.add_argument('--mask', '-m',
                      default=_default_mask,
                      help='Events to watch for')
    opts.add_argument('dir',
                      nargs='+',
                      type=Path,
                      help='Watch for events in given dir')

    args = opts.parse_args()

    flags = (Mask[tok.upper()] for tok in _splitter.split(args.mask))
    mask = reduce(or_, flags)

    args.func(args.cls, args.dir, mask)


if __name__ == '__main__':
    _main()



from typing import Dict, Tuple
from pathlib import Path
from abc import ABC, abstractmethod

import asyncio

from ._mask import Mask
from ._event import Event
from ._base import InotifyBase

__all__ = ('Minotaur',)


class Notification(ABC):
    __slots__ = (
        '_path',
        '_type',
        '_isdir',
        '_unmount',
        '_qoverflow',
    )

    def __init__(self,
                 path: Path,
                 type: Mask,
                 isdir: bool,
                 unmount: bool,
                 qoverflow: bool = False):
        self._path = path
        self._type = type
        self._isdir = bool(isdir)
        self._unmount = bool(unmount)
        self._qoverflow = bool(qoverflow)

    @property
    def isdir(self) -> bool:
        return self._isdir

    @property
    def unmount(self) -> bool:
        return self._unmount

    @property
    def qoverflow(self) -> bool:
        return self._qoverflow

    @property
    def path(self) -> Path:
        return self._path

    def __repr__(self):
        t = self._isdir and 'dir' or 'file'
        return f'{type(self).__name__}({self._type.name} {t} {self._path})'

    @classmethod
    def create(cls, path: Path, mask: Mask):
        return cls(path,
                   mask & Mask.EVENT_TYPE,
                   bool(mask & Mask.ISDIR),
                   bool(mask & Mask.UNMOUNT),
                   bool(mask & Mask.Q_OVERFLOW))


class Minotaur(InotifyBase):
    """
    Fancy interface for Inotify which does questionable things like:

    1. Resolve watch-descriptors back to paths (which races with renames of
       original paths and can't be used safely, but other inotify packages
       provide this feature, so here it is for your delectation).
    2. Link rename_from/rename_to events together. This feature would be
       useful but isn't yet actually implemented. Working on it...
    """

    __slots__ = (
        '_wdmap',
        '_cmap',
    )

    _wdmap: Dict[int, Path]
    _cmap: Dict[Tuple[int, int], Event]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._wdmap = {}
        self._cmap = {}

    def add_watch(self, p: Path, mask: Mask) -> int:
        try:
            wd = super().add_watch(p, mask)
        except Exception:
            raise
        else:
            self._wdmap[wd] = p.resolve()
        return wd

    def rm_watch(self, wd: int) -> int:
        try:
            return super().rm_watch(wd)
        except Exception:
            raise
        else:
            del self._wdmap[wd]

    def _resolve_path(self, wd: int, name: Path):
        try:
            base_dir = self._wdmap[wd]
        except KeyError:
            path = name
        else:
            path = base_dir / name

        return path

    def __next__(self) -> Notification:
        evt = super()._next_event()
        if evt is None:
            raise StopIteration

        # TODO: Link rename_from/rename_to together if we have them
        path = self._resolve_path(evt.wd, evt.name)
        return Notification.create(path, evt.mask)

    async def __anext__(self) -> Notification:
        evt = await super()._next_event_async()
        if evt is None:
            raise StopIteration

        path = self._resolve_path(evt.wd, evt.name)
        return Notification.create(path, evt.mask)



#!/usr/bin/env python3

from typing import Dict, Any
from pathlib import Path
import setuptools

_c_extension = setuptools.Extension('minotaur._inotify',
                                    ['minotaur/_inotify.c',])

version_file = Path('minotaur/__version__.py')
v: Dict[str, Any] = {}
exec(version_file.read_text(), v)

setuptools.setup(
    name=v['__title__'],
    version=v['__version__'],
    description=v['__description__'],
    author=v['__author__'],
    author_email=v['__author_email__'],
    package_data={
        'minotaur': ['py.typed'],
    },
    long_description=Path('README.md').read_text(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    license=v['__license__'],
    platforms='Linux',
    packages=[
        'minotaur',
    ],
    url=v['__url__'],
    ext_modules=[_c_extension,],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)



__title__ = 'minotaur'
__description__ = 'Pythonic, asynchronous, inotify interface.'
__url__ = 'https://github.com/giannitedesco/minotaur'
__author__ = 'Gianni Tedesco'
__author_email__ = 'gianni@scaramanga.co.uk'
__copyright__ = 'Copyright 2020 Gianni Tedesco'
__license__ = 'Apache 2.0'
__version__ = '0.0.2'
